--##########################################################################
-- XAPP 219 : Transposed Form FIR Filters

-- Engineer:  Vikram Pasham
--            Xilinx Applications
-- Date :     9/4/2000

-- The following code implements a full precision Transposed Form Filter with 
-- 8 taps, 16 bit inputs and 14-bit signed coefficients. 

-- Signal COEFF_SIGN [7:0] is assigned based on the signs of the 
-- 8 tap Coeffiecients. 
-- COEFF_SIGN[7] indicates sign of TAP 7 coefficient
-- COEFF_SIGN[6] indicates sign of TAP 6 coefficient
-- COEFF_SIGN[5] indicates sign of TAP 5 coefficient
-- COEFF_SIGN[4] indicates sign of TAP 4 coefficient
-- COEFF_SIGN[3] indicates sign of TAP 3 coefficient
-- COEFF_SIGN[2] indicates sign of TAP 2 coefficient
-- COEFF_SIGN[1] indicates sign of TAP 1 coefficient
-- COEFF_SIGN[0] indicates sign of TAP 0 coefficient


--###########################################################################





library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
use work.pack_rom16x18.all;

entity FILTER is
generic(tap_size: integer := 8);
port ( RESET: in std_logic;
       CLOCK: in std_logic;
       INA: in std_logic_vector(15 downto 0);
       LOCKED: out std_logic;
      FIR_OUT: out std_logic_vector(36 downto 0));
end FILTER;


architecture RTL of FILTER is 
type bus30 is array (0 to tap_size-1) of std_logic_vector(29 downto 0);
signal COEFF_SIGN :std_logic_vector(7 downto 0);
signal CLK, CLKx4: std_logic;
signal DLL_LOCK: std_logic;
signal A_REG: std_logic_vector(15 downto 0);
signal mul0, mul1, mul2, mul3: std_logic_vector(29 downto 0);
signal mul4, mul5, mul6, mul7: std_logic_vector(29 downto 0);
signal sum7: std_logic_vector(29 downto 0);
signal sum6: std_logic_vector(30 downto 0);
signal sum5: std_logic_vector(31 downto 0);
signal sum4: std_logic_vector(32 downto 0);
signal sum3: std_logic_vector(33 downto 0);
signal sum2: std_logic_vector(34 downto 0);
signal sum1: std_logic_vector(35 downto 0);
signal sum0: std_logic_vector(36 downto 0);
signal sign_in: std_logic;
signal kcm_result: bus30;



component DLL_CLOCK 
    port (CLKIN : in std_logic;
          RESET : in  std_logic;
          CLK0  : out std_logic;
          CLKDV : out std_logic;
          LOCKED: out std_logic);
end component;

component KCM
generic(select_rom: integer);
port(RESET  : in std_logic;
       CLK  : in std_logic;
     CLKx4  : in std_logic; 
     ENABLE : in std_logic;
	 SIGN_IN : in std_logic;
	 SIGN_COEFF: in std_logic;
 Digital_in : in std_logic_vector(15 downto 0);
     MUL_OUT: out std_logic_vector(29 downto 0));
end component;


begin

COEFF_SIGN <= "11110000";

DLLmap: DLL_CLOCK port map(CLKIN => CLOCK, RESET => RESET, CLK0 => CLKx4, 
			   CLKDV=> CLK, LOCKED => DLL_LOCK);
			   
-- Instantiate eight different KCMs with different ROMs.

KCM_MULTIPLIERs: for tap in 0 to tap_size-1 generate
                

U1:   KCM   generic map(select_rom => tap)
		       port map(CLK => CLK, CLKx4 => CLKx4, RESET => reset, ENABLE => DLL_LOCK, 
	                SIGN_IN => sign_in, SIGN_COEFF => COEFF_SIGN(tap), 
                   Digital_in => A_REG, MUL_OUT => kcm_result(tap));	 
                 end generate;

LOCKED <= DLL_LOCK;


mul7 <= kcm_result(7);  
mul6 <= kcm_result(6);
mul5 <= kcm_result(5);  
mul4 <= kcm_result(4); 
mul3 <= kcm_result(3);
mul2 <= kcm_result(2);
mul1 <= kcm_result(1);
mul0 <= kcm_result(0);


  process(CLK, RESET) 
   begin

     if(RESET = '1') then
	     FIR_OUT <= (others =>'0') ; 
       sum0 <= (others => '0');
       sum1 <= (others =>'0') ; 
       sum2 <= (others =>'0') ; 
       sum3 <= (others =>'0') ; 
       sum4 <= (others =>'0') ;
       sum5 <= (others =>'0') ;
       sum6 <= (others =>'0') ;
       sum7 <= (others =>'0') ;
       A_REG <=(others => '0');
	    sign_in <= '0';
     elsif(CLK'event and CLK ='1') then
        
	   if(INA(15) = '1') then
	      A_REG <= (not INA) + '1';
		  sign_in <= '1';
	   else
	      A_REG <=  INA;
		  sign_in <= '0';
	   end if;
        
-- The precision of the filter is preserved at each tap 
      FIR_OUT <= (sum1(35) & sum1) + (mul0(29) & mul0(29) & mul0(29) & mul0(29) & mul0(29) & mul0(29) & mul0(29) & mul0);
       sum1  <=  (sum2(34) & sum2) + (mul1(29) & mul1(29) & mul1(29) & mul1(29) & mul1(29) & mul1(29) & mul1);
       sum2  <=  (sum3(33) & sum3) + (mul2(29) & mul2(29) & mul2(29) & mul2(29) & mul2(29) & mul2);
       sum3  <=  (sum4(32) & sum4) + (mul3(29) & mul3(29) & mul3(29) & mul3(29) & mul3);
       sum4  <=  (sum5(31) & sum5) + (mul4(29) & mul4(29) & mul4(29) & mul4);
       sum5  <=  (sum6(30) & sum6) + (mul5(29) & mul5(29) & mul5);
       sum6  <=  (sum7(29) & sum7) + (mul6(29) & mul6);
       sum7  <=   mul7;


     end if;
 end process;
end RTL;






























