--########################################################################## 
-- XAPP 219 : Transposed Form FIR filters 

-- Engineer:  Vikram Pasham 
--            Xilinx Applications
-- Date :     9/4/2000

-- The following code implements a Constant Coefficent Multiplier (KCM) with
-- 16 bit signed inputs and 14 bit coefficents


-- Signal COEFF_SIGN [7:0] is assigned based on the signs of the 
-- 8 tap Coeffiecients. 
-- COEFF_SIGN[7] indicates sign of TAP 7 coefficient
-- COEFF_SIGN[6] indicates sign of TAP 6 coefficient
-- COEFF_SIGN[5] indicates sign of TAP 5 coefficient
-- COEFF_SIGN[4] indicates sign of TAP 4 coefficient
-- COEFF_SIGN[3] indicates sign of TAP 3 coefficient
-- COEFF_SIGN[2] indicates sign of TAP 2 coefficient
-- COEFF_SIGN[1] indicates sign of TAP 1 coefficient
-- COEFF_SIGN[0] indicates sign of TAP 0 coefficient


--###########################################################################



library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
use WORK.pack_rom16x18.all;


entity KCM is
generic( select_rom: integer);
port( RESET	   	: in std_logic;
       CLK  		: in std_logic;
       CLKx4		: in std_logic;
       ENABLE		: in std_logic;
	   SIGN_IN	: in std_logic;
	   SIGN_COEFF   : in std_logic;
     DIGITAL_IN         : in std_logic_vector(15 downto 0);
     MUL_OUT		: out std_logic_vector(29 downto 0));
end KCM;

architecture KCM_ARCH of KCM is

signal INA:  std_logic_vector(15 downto 0);
signal addr, rom_addr, mul_one_lsb, mul_two_lsb, mul_three_lsb: std_logic_vector(3 downto 0);
signal count : std_logic_vector(1 downto 0) ;
signal start_count: std_logic ;
signal sign_value, sign_in_reg, sign_coeff_reg: std_logic;
signal ready, out_en, out_en_reg : std_logic;
signal sum, rom_out : std_logic_vector(17 downto 0);
signal mul_result, mul_temp, mul_temp2, mul_temp3: std_logic_vector(29  downto 0);
signal shiftsum: std_logic_vector(13 downto 0) ; 
signal rom_table: mem_array;


begin 




-- Reset and Enable logic

  process(CLK, RESET, Digital_in)
   begin
    if (RESET = '1') then
 	     start_count <= '0';
       INA <= (others => '0');
    elsif (CLK'event and CLK = '1') then
     if (ENABLE = '1') then
        start_count <= '1';
     else
        start_count <= '0';
     end if;
    end if;    

    INA <= Digital_in;
          
   end process;



-- 2-bit counter which controls the address lines for ROM
-- This counter is enabled after clock CLKDLL locks.

  process(RESET, CLKx4,start_count)
  begin
   if(RESET = '1') then
     count <= "00";
   elsif(CLKx4'event and CLKx4 = '1') then
     if(start_count = '1') then
      count <= count + '1';
     else
      count <= "00";
     end if;
   end if;
  end process; 




  process(RESET, CLKx4,count, INA, rom_addr, mul_one_lsb, mul_two_lsb, mul_three_lsb, sign_value, rom_out, sum, shiftsum, sign_coeff, ready, mul_result, mul_temp)
   variable rom_table: mem_array;
    variable partial_sum,  partial_product: std_logic_vector(17 downto 0) ;
    
--    Assign ROM values for each Constant Coefficent Multiplier (KCM)
    
	procedure assign_rom is
	 begin	
	   case select_rom is
     		when 0 => rom_table := rom_table0;      -- tap0
     		when 1 => rom_table := rom_table1;						-- tap1
     		when 2 => rom_table := rom_table2;						-- tap2
     		when 3 => rom_table := rom_table3;						-- tap3
       		when 4 => rom_table := rom_table4;						-- tap4
       		when 5 => rom_table := rom_table5;						-- tap5
       		when 6 => rom_table := rom_table6;						-- tap6
      		when 7 => rom_table := rom_table7;						-- tap7
     		when others => null;
	   end case;
	end ;

   begin

  
    if( RESET = '1') then
	    
        rom_addr <= "0000";
	mul_three_lsb <= "0000";
	mul_two_lsb   <= "0000";
	mul_one_lsb   <= "0000";
	sum <= (others => '0') ;

        ready <= '0';
        out_en_reg <= '0';

        sign_value <= '0';
        sign_in_reg <= '0';      	  
 	sign_coeff_reg <= '0';
        
        
        
    elsif (CLKx4'event and CLKx4 = '1') then
  
       rom_addr <= addr;

       ready <= out_en_reg;             
       out_en_reg <= out_en;

       sign_coeff_reg <= sign_coeff;
       sign_in_reg <= sign_in;
       sign_value <= sign_in_reg xor sign_coeff_reg;     -- Sign magnitude of the Multiplier output       

       mul_one_lsb <= mul_two_lsb;          -- [2:0]- Three LSB bits of the KCM output
       mul_two_lsb <= mul_three_lsb;        -- [5:3]- Next three LSB bits of the KCM output
       mul_three_lsb <= sum(3 downto 0);    -- [8:6]- Next three LSB bits of the KCM output
       sum <= partial_sum;
       
    end if;


    case count is
       when "00" => addr <= INA(3 downto 0);
                            out_en  <= '0';
       when "01" => addr <= INA(7 downto 4);
                            out_en  <= '0' ;
       when "10" => addr <= INA(11 downto 8);
 											out_en  <= '0' ;										
       when "11" => addr <= INA(15 downto 12);
											out_en <= '1';
       when others => null;
     end case; 


    -- Read the ROM content at address "rom_addr"
                                               
    assign_rom;
    rom_out <= conv_std_logic_vector(rom_table(conv_integer(rom_addr)), 18);
    

    if(SIGN_COEFF = '1') then                -- Check the sign of the ROM output
           partial_product := not (rom_out) + '1'; 
    else  
         partial_product  := rom_out;
    end if;
   

   -- Shift and add the partial products

     partial_sum := partial_product + ("0000" & shiftsum);
 
     mul_temp  <= sum & mul_three_lsb & mul_two_lsb & mul_one_lsb;
     

      if(ready = '1') then
       shiftsum <= (others => '0');
      else 
       shiftsum <= sum(17 downto 4);
      end if;


      if(sign_value = '1') then               -- Adjust the mutiplier output based
        mul_result   <= not(mul_temp) + '1';  -- on input and coefficient signs. 
      elsif (sign_value = '0') then
         mul_result <= mul_temp;
      end if; 

   end process;



  process(reset, clkx4, ready, mul_result)
  begin
     if(reset = '1') then
           MUL_OUT <= (others => '0');
     elsif (clkx4'event and clkx4 = '0') then
		   if(ready = '1') then
           MUL_OUT <= mul_result;
        end if;
     end if;
   
  end process;
  
      
end KCM_ARCH;








































































































